﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

[RequireComponent( typeof( AudioSource ) )]
public class GameStateManager : MonoBehaviour
{
    public static GameStateManager Instance;

    public Target TargetPrefab;
    public Transform[] TargetEmitters;
    public Text ScoreText;
    public RectTransform[] StrikeMarks;

    private AudioSource failAudioSource;
    private bool gameRunning;
    private int score;
    private int strikes;

    void Start()
    {
        // Save ourselves as the singleton instance:
        GameStateManager.Instance = this;

        this.failAudioSource = this.GetComponent<AudioSource>();
    }

    void Update()
    {
        // Update the scoreboard:
        if ( this.gameRunning )
        {
            this.ScoreText.text = string.Format( "SCORE: {0:000}", this.score );
        }
        else
        {
            this.ScoreText.text = "GAME OVER";
        }

        for ( int i = 0; i < this.StrikeMarks.Length; i++ )
        {
            this.StrikeMarks[i].gameObject.SetActive( this.strikes > i );
        }
    }

    public void StartNewGame()
    {
        if ( !this.gameRunning )
        {
            this.gameRunning = true;
            this.StartCoroutine( this.Script() );
        }
    }

    public void IncreaseScore()
    {
        this.score += 1;
    }

    private Target SpawnTarget( bool destroyedBySword )
    {
        Transform emitter = this.TargetEmitters[Random.Range( 0, this.TargetEmitters.Length )];
        Target target = GameObject.Instantiate<Target>( this.TargetPrefab );
        target.Initialize( emitter.position, emitter.up, destroyedBySword );
        return target;
    }

    IEnumerator Script()
    {
        // Reset the game state:
        this.score = 0;
        this.strikes = 0;

        // Pause for a few seconds to let the player get situated:
        yield return new WaitForSeconds( 3.0f );

        // Spawn waves of targets until the player gets three strikes:
        int roundNumber = 0;
        Queue<bool> spawnQueue = new Queue<bool>();
        while ( true )
        {
            // Figure out what to spawn:
            if ( roundNumber == 0 )
            {
                spawnQueue.Enqueue( true );
            }
            else if ( roundNumber == 1 )
            {
                spawnQueue.Enqueue( false );
            }
            else
            {
                for ( int i = 0; i < ( roundNumber - 2 ) / 4 + 2; i++ )
                {
                    spawnQueue.Enqueue( Random.Range( 0, 2 ) == 0 );
                }
            }

            // Spawn the targets with brief delays between:
            List<Target> spawnedTargets = new List<Target>();
            while ( spawnQueue.Count > 0 )
            {
                spawnedTargets.Add( this.SpawnTarget( spawnQueue.Dequeue() ) );
                yield return new WaitForSeconds( 0.3f );
            }

            // Give the player a few seconds to destroy everything:
            yield return new WaitForSeconds( 2.75f );

            // If the player failed to destroy everything, give them a strike, possibly ending the game:
            if ( spawnedTargets.Any( x => x != null ) )
            {
                this.strikes += 1;
                this.failAudioSource.Play();
                yield return new WaitForSeconds( 1.5f );

                if ( this.strikes >= this.StrikeMarks.Length )
                {
                    yield return new WaitForSeconds( 1.0f );
                    this.gameRunning = false;
                    yield break;
                }
            }

            // Destroy all the targets we created:
            foreach ( Target target in spawnedTargets )
            {
                if ( target != null )
                {
                    GameObject.Destroy( target.gameObject );
                }
            }

            // Advance to the next round:
            roundNumber += 1;
        }
    }
}
